from qtpy import QtWidgets
from pymodaq.control_modules.viewer_utility_classes import DAQ_Viewer_base, comon_parameters, main

import numpy as np
from collections import OrderedDict
from pymodaq.utils.daq_utils import ThreadCommand, getLineInfo
from pymodaq.utils.data import DataFromPlugins, Axis, DataToExport
import sys
import clr
from easydict import EasyDict as edict

from pymodaq_plugins_oceaninsight.utils import Config

plugin_config = Config()

omnidriver_path = plugin_config('omnidriver_path')
try:
    sys.path.append(omnidriver_path)
    clr.AddReference("NETOmniDriver-NET40")
    import OmniDriver as omnidriver

except:
    omnidriver = None


class DAQ_1DViewer_Omnidriver(DAQ_Viewer_base):
    """PyMoDAQ plugin controlling spectrometers from OceanOptics and using their OmniDriver wrapper ('NETOmniDriver-NET40')
    The wrapper is interfaced using the clr package and then importing Omnidriver as a standart python package
    https://oceanoptics.com/api/omnidriver/index.html?overview-summary.html

    The plugin enables the discovery of any connected usb spectrometers and can control them in parallel. It extracts the
    calibrated wavelength and will export datas that will be plotted with respect to this wavelengths vector.

    Used methods from the wrapper:
    ------------------------------
     setIntegrationTime
     getIntegrationTime

    See Also
    --------
    utility_classes.DAQ_Viewer_base
    """

    params = comon_parameters + [
        {'title': 'N spectrometers:', 'name': 'Nspectrometers', 'type': 'int', 'value': 0,
         'default': 0, 'min': 0, 'readonly': True},
        {'title': 'Spectrometers:', 'name': 'spectrometers', 'type': 'group', 'children': []},
        ]

    hardware_averaging = True

    def ini_attributes(self):

        self.controller: omnidriver.NETWrapper = None
        self.spectro_names = [] #contains the spectro name as returned from the wrapper
        self.spectro_id = [] # contains the spectro id as set by the ini_detector method and equal to the Parameter name


    def commit_settings(self, param):
        """
        """
        if param.name() == 'exposure_time':
            ind_spectro = self.spectro_id.index(param.parent().name())
            self.controller.setIntegrationTime(ind_spectro, param.value() * 1000)

            param.setValue(self.controller.getIntegrationTime(ind_spectro) / 1000)

    def ini_detector(self, controller=None):
        """
            Initialisation procedure of the detector updating the status dictionnary.

            See Also
            --------
            set_Mock_data, daq_utils.ThreadCommand
        """
        self.ini_detector_init(controller, None)

        if self.is_master:
            self.controller = omnidriver.NETWrapper()

            N = self.controller.openAllSpectrometers()
            self.settings.child('Nspectrometers').setValue(N)
            self.spectro_names = []
            self.spectro_id = []

            data_init = DataToExport('Spectro')

            for ind_spectro in range(N):
                name = self.controller.getName(ind_spectro)
                self.spectro_names.append(name)
                self.spectro_id.append('spectro{:d}'.format(ind_spectro))

                exp_max = self.controller.getMaximumIntegrationTime(ind_spectro)
                exp_min = self.controller.getMinimumIntegrationTime(ind_spectro)
                exp = self.controller.getIntegrationTime(ind_spectro) / 1000
                wavelengths = self.get_xaxis(ind_spectro)

                data_init.append(DataFromPlugins(name=name, data=[np.zeros_like(wavelengths)],
                                                 dim='Data1D',
                                                 x_axis=Axis(data=wavelengths, label='Wavelength',
                                                             units='nm')))

                for ind in range(2): #this is to take into account that adding it once doen't work (see pyqtgraph Parameter...)
                    try:
                        self.settings.child('spectrometers').addChild(
                            {'title': name,'name': 'spectro{:d}'.format(ind_spectro),
                             'type': 'group', 'children':[
                                {'title': 'grab spectrum:','name': 'grab', 'type': 'bool',
                                 'value': True},
                            {'title': 'Exposure time (ms):','name': 'exposure_time', 'type': 'int',
                             'value': int(exp), 'min': int(exp_min/1000), 'max': int(exp_max/1000)},

                            ]
                             })
                    except:
                        pass

                QtWidgets.QApplication.processEvents()

            #init viewers

            if N == 0:
                raise Exception('No detected hardware')
            self.dte_signal_temp.emit(data_init)

        initialized = True
        info = 'Ocean Insight Spectrometers initialized'
        return info, initialized

    def get_xaxis(self, ind_spectro):
        wavelengths_chelou = self.controller.getWavelengths(ind_spectro)
        wavelengths = np.array([wavelengths_chelou[ind] for ind in range(len(wavelengths_chelou))])

        return wavelengths

    def close(self):
        """
        """
        if self.controller is not None:
            self.controller.closeAllSpectrometers()

    def grab_data(self, Naverage=1, **kwargs):
        """

        """
        try:
            dte = DataToExport('Spectro')
            for ind_spectro in range(len(self.spectro_names)):
                if self.settings.child('spectrometers', 'spectro{:d}'.format(ind_spectro), 'grab').value():
                    self.controller.setScansToAverage(ind_spectro, Naverage)
                    data_chelou = self.controller.getSpectrum(ind_spectro)
                    data_array = np.array([data_chelou[ind] for ind in range(len(data_chelou))])
                    dte.append(
                      DataFromPlugins(name=self.spectro_names[ind_spectro],
                                      data=[data_array], dim='Data1D',
                                      axes=[Axis('wavelength',
                                                 units='m',
                                                 data=self.get_xaxis(ind_spectro)
                                                      * 1e-9)]))
                    QtWidgets.QApplication.processEvents()

            self.dte_signal.emit(dte)


        except Exception as e:
            self.emit_status(ThreadCommand('Update_Status', [getLineInfo() + str(e), "log"]))

    def stop(self):
        """

        """
        for ind_spec, name in enumerate(self.spectro_names):
            self.controller.stopAveraging(ind_spec)


if __name__ == '__main__':
    main(__file__, init=False)



import numpy as np
from easydict import EasyDict as edict
from pymodaq.utils.daq_utils import ThreadCommand, getLineInfo
from pymodaq.utils.data import DataFromPlugins, Axis, DataToExport
from pymodaq.control_modules.viewer_utility_classes import DAQ_Viewer_base, comon_parameters, main
from timeit import timeit

# explicitly request cseabreeze
import seabreeze
seabreeze.use('cseabreeze')
from seabreeze.spectrometers import Spectrometer, list_devices


class DAQ_1DViewer_Seabreeze(DAQ_Viewer_base):
    """
    """
    # Upon initialisation
    devices = list_devices()

    #Supports pseudo hardware-averaging
    hardware_averaging = True

    params = comon_parameters + [
        {'title': 'Device:', 'name': 'device', 'type': 'list', 'limits': devices},
        {'title': 'Integration (ms):', 'name': 'integration', 'type': 'float', 'value': 1.0},
        {'title': 'Advanced:', 'name': 'advanced', 'type': 'group', 'children': [
            {'title': 'Serial Number:', 'name': 'serial_number', 'type': 'str', 'value': "", 'readonly': True},
            {'title': 'Non Linearity correction:', 'name': 'correct_non_linearity', 'type': 'bool', 'value': False},
            {'title': 'Max Intensity', 'name': 'max_intensity', 'type': "float", 'value': 65535, 'readonly': True},
            {'title': 'Pixels:', 'name': 'pixels', 'type': 'int', 'value': 2048, 'readonly': True},
            {'title': 'Dark Channels:', 'name': 'dark_channels', 'type': 'int', 'value': 10, 'readonly': True},
            {'title': 'Readout Time (ms)', 'name': 'readout_time', 'type': 'float', 'value': 666, 'readonly': True},
        ]}
    ]

    def ini_attributes(self):
        self.controller: Spectrometer = None

    def commit_settings(self, param):
        """
        """
        if param.name() == "integration":
            self.controller.integration_time_micros(param.value() * 1000)
            # There is no way to get feedback from the spectrometer on the currently set integration time
        # elif ...
        ##

    def ini_detector(self, controller=None):
        """Detector communication initialization

        Parameters
        ----------
        controller: (object) custom object of a PyMoDAQ plugin (Slave case). None if only one detector by controller (Master case)

        Returns
        -------
        self.status (edict): with initialization status: three fields:
            * info (str)
            * controller (object) initialized controller
            *initialized: (bool): False if initialization failed otherwise True
        """
        self.ini_detector_init(controller, None)

        if self.is_master:
            # From what I understand this will just get the spectro currently selected in the list
            dvc = self.settings['device']
            self.controller = Spectrometer(dvc)
            self.controller.open()
            #####################################

        # Oceanoptics spectrometers (at least the ones i Know) have fixed axis
        # get inactive pixels
        c0 = self.controller.f.spectrometer.get_electric_dark_pixel_indices()[-1]
        self.settings.child('advanced').child('dark_channels').setValue(c0)
        # get the x_axis
        data_x_axis = self.controller.wavelengths()  # Way to get the x axis
        data_x_axis = data_x_axis[c0:] * 1e9  # Get rid of the dark pixels
        self.x_axis = Axis(data=data_x_axis, label='wavelength', units='m', index=0)

        # Get the name
        specname = f"Ocean Insight {self.controller.model}"

        # initialize viewers pannel with the future type of data

        self.data_grabed_signal_temp.emit([
            DataFromPlugins(name=specname,
                            data=[np.zeros_like(data_x_axis)],
                            dim='Data1D',
                            labels=['Intensity'],
                            axes=[self.x_axis]), ])


        # Update the parameters
        # Here we need to do a few things. Get the integration time limits and set it in the settings
        tlimits = np.array(self.controller.integration_time_micros_limits) / 1000
        self.settings.child('integration').setLimits(tlimits)
        # Here we need to update the advanced parameters
        advanced_settings = self.settings.child('advanced')
        sn = self.controller.serial_number
        advanced_settings.child('serial_number').setValue(sn)
        # non linearity coefficients
        nlc_feat = self.controller.f.nonlinearity_coefficients
        if nlc_feat is not None and not any(np.isnan(nlc_feat.get_nonlinearity_coefficients())):
            advanced_settings.child('correct_non_linearity').setValue(True)
        else:
            advanced_settings.child('correct_non_linearity').setValue(False)
            advanced_settings.child('correct_non_linearity').setOpts(enabled=False)

        # measure the readout time
        Nperf = 200
        self.settings.child('integration').setValue(tlimits[0])  # Update the parameter value to the lower limit
        self.controller.integration_time_micros(tlimits[0]*1000)  # Set the readout time to lower limit
        perf = timeit(lambda: self.controller.intensities(), number=Nperf)  # time the execution of code in [s]
        self.settings.child('advanced').child('readout_time').setValue(1000*perf/Nperf)  # set the settings

        ##############################

        info = f"Initialized {specname} spectrometer"
        initialized = True
        return info, initialized

    def close(self):
        """
        Terminate the communication protocol
        """
        if self.controller is not None:
            self.controller.close()

    def grab_data(self, Naverage = 1, **kwargs):
        """
        Parameters
        ----------
        Naverage: (int) Number of hardware averaging
        kwargs: (dict) of others optionals arguments
        """
        nlc = self.settings.child('advanced').child('correct_non_linearity').value()
        c0 = self.settings.child('advanced').child('dark_channels').value()
        # synchrone version (blocking function)
        # Pseudo-hardware-averaging
        if Naverage > 1:
            data = [self.controller.intensities(correct_nonlinearity=nlc)[c0:] for i in range(Naverage)]
            data = np.array(data)
            data = data.mean(0)
        # Otherwise normal single-acquisition
        else:
            data = self.controller.intensities(correct_nonlinearity=nlc)[c0:]

        self.dte_signal.emit(DataToExport('Spectro', data=[
            DataFromPlugins(name='oceanseabreeze', data=[data], dim='Data1D',
                            labels=['spectrum'], axes=[self.x_axis.copy()])]))

    def stop(self):
        pass


if __name__ == '__main__':
    main(__file__, init=False)
